"""Set up tests."""

from negate import Negator


# Reuse negator model for all tests.
negator_model: Negator = None


def pytest_addoption(parser):
    parser.addoption(
        "--transformers", action="store_true", help="use Transformers")


def pytest_generate_tests(metafunc):
    if "negator" in metafunc.fixturenames:
        # Initialize negator only once.
        global negator_model
        if negator_model is None:
            use_transformers = metafunc.config.getoption("transformers")
            try:
                # `use_gpu` ignored if `use_transformers` is False.
                negator_model = Negator(
                    use_transformers=use_transformers, use_gpu=True)
            except ValueError:  # "No GPU devices detected"
                negator_model = Negator(
                    use_transformers=use_transformers, use_gpu=False)
        metafunc.parametrize("negator", [negator_model])



"""Negation tools."""

import logging
import importlib
import os
import sys
import spacy
from contextlib import contextmanager
from typing import Dict, List, Optional, Tuple, Union
from lemminflect import getInflection, getLemma
from spacy.symbols import AUX, neg, VERB
from spacy.tokens import Doc as SpacyDoc
from spacy.tokens import Token as SpacyToken
from .tokens import Token


class Negator:
    """Negator for the English language."""

    def __init__(
        self,
        use_transformers: Optional[bool] = None,
        use_gpu: Optional[bool] = None,
        fail_on_unsupported: Optional[bool] = None,
        log_level: Optional[int] = None
    ):
        """Instanciate a :obj:`Negator`.

        Args:
            use_transformers (:obj:`Optional[bool]`, defaults to :obj:`False`):
                Whether to use a Transformer model for POS tagging and
                dependency parsing.

                .. note::

                   When set to :obj:`True` the model <en_core_web_trf
                   `https://spacy.io/models/en#en_core_web_trf`>__ is used.
            use_gpu (:obj:`Optional[bool]`, defaults to :obj:`False`):
                Whether to use the GPU, if available. This parameter is ignored
                when :param:`use_transformers` is set to :obj:`False`.
            fail_on_unsupported (:obj:`Optional[bool]`, defaults to :obj:`False`):
                Whether to fail upon non-supported sentences. If set to
                :obj:`False`, a warning will be printed, and the sentence will
                try to be negated in a best-effort fashion.
            log_level (:obj:`Optional[int]`, defaults to ``logging.INFO``):
                The level of the logger.

        Raises:
            :obj:`RuntimeError`: If the sentence is not supported and
            :arg:`fail_on_unsupported` is set to :obj:`True`.
        """
        if use_transformers is None:
            use_transformers = False
        if use_gpu is None:
            use_gpu = False
        if fail_on_unsupported is None:
            fail_on_unsupported = False
        if log_level is None:
            log_level = logging.INFO

        # Set up logger.
        logging.basicConfig(
            format="%(name)s - %(levelname)s: %(message)s",
            level=log_level
        )
        self.logger = logging.getLogger(__class__.__name__)
        self.fail_on_unsupported = fail_on_unsupported
        # Load spaCy model. If not available locally, the model will be first
        # installed.
        if use_transformers and use_gpu:
            spacy.require_gpu()
        self.spacy_model = self._initialize_spacy_model(use_transformers)
        # Initialize AUX negation dictionary.
        self._initialize_aux_negations()
        # Store whether tokens have a whitespace after them. This is used later
        # on for detokenization.
        SpacyToken.set_extension("has_space_after", default=True, force=True)

    def negate_sentence(
        self,
        sentence: str,
        prefer_contractions: bool = True
    ) -> str:
        if not sentence:
            return ""
        doc = self._parse(sentence)
        root = self._get_entry_point(doc)

        if not root or not self._is_sentence_supported(doc):
            self._handle_unsupported()
            if not root:  # Don't even bother trying :)
                return sentence

        # Any negations we can remove? (e.g.: "I don't know.", "They won't
        # complain.", "He has not done it.", etc.).
        negation = self._get_negated_child(root)
        # Handle sentences such as "Not to mention that they weren't there."
        # We want to prevent taking the first "Not" as the negation, since this
        # complicates things.
        first_aux_or_verb = self._get_first_aux_or_verb(doc)
        while (negation and first_aux_or_verb
                   and negation.i < first_aux_or_verb.i):
            # Search for another negation, if any.
            negation = self._get_negated_child(root, min_index=negation.i+1)
        if negation:
            aux_child = self._get_aux_child(root)
            # General verbs -> Remove negation and conjugate verb.
            # If there is an AUX child, we need to "unnegate" the AUX instead.
            if (not self._is_aux(root) and root.tag_ != "VBN"
                    and not aux_child and not self._is_verb_to_do(root)
                    and not self._is_verb_to_be(root)):
                remove = [root.i, negation.i]
                add = {root.i: Token(
                    text=self.conjugate_verb(root.text, root.tag_),
                    has_space_after=negation._.has_space_after
                )}
            # Special case AUX "won't" -> Remove negation and replace
            # "wo" -> "will".
            elif root.text.lower() == "wo":
                remove = [root.i, negation.i]
                add = {root.i: Token(
                    text=" will",
                    has_space_after=negation._.has_space_after
                )}
            elif aux_child and aux_child.text.lower() == "wo":
                remove = [aux_child.i, negation.i]
                add = {aux_child.i: Token(
                    text=" will",
                    has_space_after=negation._.has_space_after
                )}
            # Special case AUX "can't" -> Remove negation and replace
            # "ca" -> "can".
            elif root.text.lower() == "ca":
                remove = [root.i, negation.i]
                add = {root.i: Token(
                    text=" can",
                    has_space_after=negation._.has_space_after
                )}
            elif aux_child and aux_child.text.lower() == "ca":
                remove = [aux_child.i, negation.i]
                add = {aux_child.i: Token(
                    text=" can",
                    has_space_after=negation._.has_space_after
                )}
            # Any other AUX or verb in past participle -> Remove negation.
            else:
                remove = [root.i, negation.i]
                # Correctly handle space in e.g., "He hasn't been doing great."
                if negation.i < root.i and negation.i > 0:
                    doc[negation.i-1]._.has_space_after = negation._.has_space_after
                # Correctly handle space in e.g., "I'm not doing great." vs.
                # "I am not doing great."
                space_before = " " * int(root.i > 0
                                        and doc[root.i-1]._.has_space_after)
                # Negation can come before ("She will not ever go.") or after
                # the root ("She will not."). Space after is different in each
                # case.
                space_after = (negation._.has_space_after if negation.i > root.i
                               else root._.has_space_after)
                add = {root.i: Token(
                    text=f"{space_before}{root.text}",
                    has_space_after=space_after
                )}
                # Special case "do" -> Also remove "do" and conjugate verb.
                # E.g.: "He doesn't like it." -> "He likes it".
                if aux_child and self._is_verb_to_do(aux_child):
                    remove.append(aux_child.i)
                    add[root.i] = Token(
                        text=f"{space_before}"
                             f"{self.conjugate_verb(root.text.lower(), aux_child.tag_)}",
                        has_space_after=root._.has_space_after
                    )
            return self._compile_sentence(
                doc,
                remove_tokens=remove,
                add_tokens=add
            )

        # AUX as ROOT (e.g.: "I'm excited.") or ROOT children e.g.,
        # "I do think...".
        if (self._is_aux(root) or self._is_verb_to_be(root)
                or any(self._is_aux(child) for child in root.children)):
            # If the AUX has AUX children, negate them instead. E.g.: "I will
            # be there." or "They have been moody lately."
            aux_child = self._get_aux_child(root)
            return self._negate_aux_in_doc(
                aux=root if not aux_child else aux_child,
                doc=doc,
                prefer_contractions=prefer_contractions
            )

        # General verb non-negated.
        negated_aux = self.negate_aux(self.conjugate_verb('do', root.tag_),
                                      prefer_contractions)
        return self._compile_sentence(
            doc,
            remove_tokens=[root.i],
            add_tokens={root.i: Token(
                text=f"{negated_aux if negated_aux is not None else ''} "
                     f"{self.get_base_verb(root.text.lower())}",
                has_space_after=root._.has_space_after
            )}
        )

    def negate_verb(
        self,
        verb: str,
        tag: str = None,
        prefer_contractions: bool = True
    ) -> str:
        if not verb:
            return ""
        negated_aux = self.negate_aux(verb, prefer_contractions)
        if negated_aux:
            return negated_aux
        if tag is None:  # infer tag
            tag = self._parse(verb)[0].tag_
        return (
            f"{self.negate_aux(self.conjugate_verb('do', tag), prefer_contractions)} "
            f"{self.get_base_verb(verb.lower())}"
        )

    def conjugate_verb(self, verb: str, tag: str) -> str:
        conjugated_verb: Tuple[str] = getInflection(verb, tag)
        return conjugated_verb[0] if conjugated_verb else verb

    def get_base_verb(self, verb: str) -> str:
        base_verb: Tuple[str] = getLemma(verb, upos="VERB")
        return base_verb[0] if base_verb else verb

    def negate_aux(
        self,
        auxiliary_verb: str,
        prefer_contractions: bool = True
    ) -> Optional[str]:
        negated_aux = self._aux_negations[prefer_contractions].get(
            auxiliary_verb
        )
        if negated_aux is None:
            self._handle_unsupported()
        return negated_aux

    def _negate_aux_in_doc(
        self,
        aux: Union[Token, SpacyToken],
        doc: SpacyDoc,
        prefer_contractions: bool = True
    ) -> str:
        negation = self._get_negated_child(aux)
        # If AUX negated -> Remove negation.
        if negation:
            # Special case AUX "won't" -> Remove negation and replace
            # "wo" -> "will".
            if aux.text.lower() == "wo":
                replace_aux = Token(
                    text=" will",
                    has_space_after=negation._.has_space_after
                )
            # Special case AUX "can't" -> Remove negation and replace
            # "ca" -> "can".
            elif aux.text.lower() == "ca":
                replace_aux = Token(
                    text=" can",
                    has_space_after=negation._.has_space_after
                )
            # Nothing to do.
            else:
                replace_aux = Token(
                    text=aux.text,
                    has_space_after=negation._.has_space_after
                )
            return self._compile_sentence(
                doc,
                remove_tokens=[aux.i, negation.i],
                add_tokens={
                    aux.i: replace_aux
                }
            )

        # If AUX not negated -> Replace AUX with its negated.
        # Disambiguate "to be" and "to have" when AUX is "'s".
        aux_text = aux.text
        if aux.text.lower() == "'s":
            parent = self._get_parent(aux, doc)
            if parent and (parent.tag_ == "VBN"
                    or any(child.tag_ == "VBN" for child in parent.children)):
                # "'s" is "to have"
                aux_text = f"{aux.text}_"
        remove = [aux.i]
        add = {
            aux.i: Token(
                text=self.negate_aux(aux_text, prefer_contractions),
                has_space_after=aux._.has_space_after
            )
        }
        # Handle e.g., "should've" -> "shouldn't have"
        if aux.i+1 < len(doc) and doc[aux.i+1].text.lower() == "'ve":
            remove.append(aux.i+1)
            add[aux.i+1] = Token(
                text=" have",
                has_space_after=doc[aux.i+1]._.has_space_after
            )
        return self._compile_sentence(
            doc,
            remove_tokens=remove,
            add_tokens=add
        )

    def _parse(self, string_: str) -> SpacyDoc:
        # Remove extra whitespaces and other non-printable chars.
        string_ = self._remove_extra_whitespaces(string_)
        # Tokenize.
        doc = self.spacy_model(string_)
        i = 0  # Used to determine whitespaces.
        for tk in doc:
            has_space_after: bool = (
                i+len(tk) < len(string_) and (string_[i+len(tk)] == " ")
            )
            tk._.has_space_after = has_space_after
            i += len(tk) + int(has_space_after)
        return doc

    def _get_entry_point(self, doc: SpacyDoc) -> Optional[SpacyToken]:
        root = self._get_root(doc)
        if root is None:  # nothing we can do
            return None
        # If the root token is not an AUX or a VERB, look for an AUX or VERB in
        # its children.
        if not (self._is_aux(root) or self._is_verb(root)):
            entry_point = None
            if root.children:
                entry_point = [tk for tk in root.children
                               if self._is_aux(tk) or self._is_verb(tk)]
            # No AUX or VERB found in the root children -> Take the first AUX
            # or VERB in the sentence, if any.
            if not entry_point:
                entry_point = [tk for tk in doc
                               if self._is_aux(tk) or self._is_verb(tk)]
            return entry_point[0] if entry_point else None
        return root

    def _get_root(self, doc: SpacyDoc) -> Optional[SpacyToken]:
        root = [tk for tk in doc if tk.dep_ == "ROOT"]
        return root[0] if root else None

    def _get_negated_child(
        self,
        token: SpacyToken,
        min_index: int = 0
    ) -> Optional[SpacyToken]:
        if not token:
            return None
        min_index = max(0, min_index)  # prevent negative values
        child = [child for child in token.children if child.dep == neg
                                                      and child.i >= min_index]
        return child[0] if child else None

    def _get_aux_child(
        self,
        token: SpacyToken,
        min_index: int = 0
    ) -> Optional[SpacyToken]:
        if not token:
            return None
        min_index = max(0, min_index)  # prevent negative values
        child = [child for child in token.children if self._is_aux(child)
                                                      and child.i >= min_index]
        return child[0] if child else None

    def _get_first_aux_or_verb(
        self,
        doc: SpacyDoc
    ) -> Optional[SpacyToken]:
        aux = [tk for tk in doc if self._is_aux(tk) or self._is_verb(tk)]
        return aux[0] if aux else None

    def _get_parent(
        self,
        token: SpacyToken,
        doc: SpacyDoc
    ) -> Optional[SpacyToken]:
        if not token:
            return None
        parent = [
            potential_parent
            for potential_parent in doc
            if token in potential_parent.children
        ]
        return parent[0] if parent else None

    def _is_aux(self, token: SpacyToken) -> bool:
        if not token:
            return False
        return token.pos == AUX

    def _is_verb(self, token: SpacyToken) -> bool:
        if not token:
            return False
        return token.pos == VERB

    def _is_verb_to_do(self, verb: SpacyToken) -> bool:
        if not verb:
            return False
        return getLemma(verb.text.lower(), "VERB")[0] == "do"

    def _is_verb_to_be(self, verb: SpacyToken) -> bool:
        if not verb:
            return False
        return getLemma(verb.text.lower(), "VERB")[0] == "be"

    def _is_sentence_supported(self, doc: SpacyDoc) -> bool:
        return any(self._is_aux(tk) or self._is_verb(tk) for tk in doc)

    def _compile_sentence(
        self,
        doc: SpacyDoc,
        remove_tokens: List[int] = None,  # indexes to remove
        add_tokens: Dict[int, Token] = None  # idx, string to add
    ) -> str:
        if remove_tokens is None:
            remove_tokens = []
        if add_tokens is None:
            add_tokens = {}
        else:
            add_tokens = dict(sorted(add_tokens.items()))  # sort by index
        tokens = [Token(tk.text, tk._.has_space_after) for tk in doc]
        for i in remove_tokens:
            tokens[i] = Token(text="", has_space_after=False)
        for count, item in enumerate(add_tokens.items()):
            i, tk = item
            tokens.insert(i+count, tk)
        return self._capitalize_first_letter(
            self._remove_extra_whitespaces(
                "".join([f"{tk.text}{' '*int(tk.has_space_after)}"
                         for tk in tokens])
                )
            )

    def _capitalize_first_letter(self, string_: str) -> str:
        if not string_:
            return ""
        return f"{string_[0].upper()}{string_[1:]}"

    def _remove_extra_whitespaces(self, string_: str) -> str:
        if not string_:
            return ""
        return " ".join(string_.split())

    def _initialize_spacy_model(
        self,
        use_transformers: bool,
        **kwargs
    ) -> spacy.language.Language:
        """Initialize the spaCy model to be used by the Negator.

        Heavily inspired by `https://github.com/BramVanroy/spacy_download`__.

        Args:
            use_transformers (:obj:`Optional[bool]`, defaults to :obj:`False`):
                Whether to use a Transformer model for POS tagging and
                dependency parsing.

                .. note::

                   When set to :obj:`True` the model <en_core_web_trf
                   `https://spacy.io/models/en#en_core_web_trf`>__ is used.
            **kwargs:
                Additional arguments passed to :func:`spacy.load`.

        Returns:
            :obj:`spacy.language.Language`: The loaded spaCy model, ready to
            use.
        """
        # See https://stackoverflow.com/a/25061573/14683209
        # We don't want the messages coming from pip to pollute the stdout.
        @contextmanager
        def suppress_stdout():
            with open(os.devnull, "w") as devnull:
                old_stdout = sys.stdout
                sys.stdout = devnull
                try:
                    yield
                finally:
                    sys.stdout = old_stdout

        model_name = "en_core_web_trf" if use_transformers else "en_core_web_md"
        try:  # Model installed?
            model_module = importlib.import_module(model_name)
        except ModuleNotFoundError:  # Download and install model.
            self.logger.info("Downloading model. This only needs to happen "
                             "once. Please, be patient...")
            with suppress_stdout():
                spacy.cli.download(model_name, False, False, "-q")
            model_module = importlib.import_module(model_name)
        return model_module.load(**kwargs)

    def _handle_unsupported(self):
        """Handle behavior upon unsupported sentences.

        Raises:
            :obj:`RuntimeError`: If :arg:`fail_on_unsupported` is set to
            :obj:`True`.
        """
        if self.fail_on_unsupported:
            raise RuntimeError("Sentence not supported.")
        else:
            self.logger.warning("Sentence not supported. Output might be "
                                "arbitrary.")

    def _initialize_aux_negations(self):
        self._aux_negations = {
            True: {
                "'m": "'m not",
                "am": "am not",
                "'re": " aren't",
                "are": "aren't",
                "'s": " isn't",
                "is": "isn't",
                "was": "wasn't",
                "were": "weren't",
                "do": "don't",
                "does": "doesn't",
                "did": "didn't",
                "'ve": " haven't",
                "have": "haven't",
                "'s_": " hasn't",  # underscore to distinguish from verb to be.
                "has": "hasn't",
                "'d": " hadn't",
                "had": "hadn't",
                "can": "can't",
                "could": "couldn't",
                "must": "mustn't",
                "might": "mightn't",
                "may": "may not",
                "should": "shouldn't",
                "ought": "oughtn't",
                "'ll": " won't",
                "will": "won't",
                "would": "wouldn't"
            },
            False: {
                "'m": " am not",
                "am": "am not",
                "'re": " are not",
                "are": "are not",
                "'s": " is not",
                "is": "is not",
                "was": "was not",
                "were": "were not",
                "do": "do not",
                "does": "does not",
                "did": "did not",
                "'ve": " have not",
                "have": "have not",
                "'s_": " has not",  # underscore to distinguish from verb to be.
                "has": "has not",
                "'d": " had not",
                "had": "had not",
                "can": "cannot",
                "could": "could not",
                "must": "must not",
                "might": "might not",
                "may": "may not",
                "should": "should not",
                "ought": "ought not",
                "'ll": " will not",
                "will": "will not",
                "would": "would not"
            }
        }



"""Data for tests.

Format::

   (input_sentence, output_sentence, prefer_contractions)
"""

# AUX as ROOT - Affirmative
aux_root_affirmative = [
    ("I should.", "I shouldn't.", True),
    ("I should.", "I should not.", False),
    ("I can.", "I can't.", True),
    ("I can.", "I cannot.", False),
    ("I will.", "I won't.", True),
    ("I will.", "I will not.", False),
    ("I'm excited.", "I'm not excited.", True),
    ("I'm excited.", "I am not excited.", False),
    ("I am excited.", "I am not excited.", True),
    ("I am excited.", "I am not excited.", False),
    ("I've been excited.", "I haven't been excited.", True),
    ("I've been excited.", "I have not been excited.", False),
    ("I have been excited.", "I haven't been excited.", True),
    ("I have been excited.", "I have not been excited.", False),
    ("I'll be excited.", "I won't be excited.", True),
    ("I'll be excited.", "I will not be excited.", False),
    ("I will be excited.", "I won't be excited.", True),
    ("I will be excited.", "I will not be excited.", False),
    ("She's excited.", "She isn't excited.", True),
    ("She's excited.", "She is not excited.", False),
    ("She is excited.", "She isn't excited.", True),
    ("She is excited.", "She is not excited.", False),
    ("She's been excited.", "She hasn't been excited.", True),
    ("She's been excited.", "She has not been excited.", False),
    ("She has been excited.", "She hasn't been excited.", True),
    ("She has been excited.", "She has not been excited.", False),
    ("You're excited.", "You aren't excited.", True),
    ("You're excited.", "You are not excited.", False),
    ("You are excited.", "You aren't excited.", True),
    ("You are excited.", "You are not excited.", False),
    ("We're excited.", "We aren't excited.", True),
    ("We're excited.", "We are not excited.", False),
    ("We are excited.", "We aren't excited.", True),
    ("We are excited.", "We are not excited.", False),
    ("They're excited.", "They aren't excited.", True),
    ("They're excited.", "They are not excited.", False),
    ("They are excited.", "They aren't excited.", True),
    ("They are excited.", "They are not excited.", False),
]

# AUX as ROOT - Negative
aux_root_negative = [
    ("I don't.", "I do.", True),
    ("I don't.", "I do.", False),
    ("I do not.", "I do.", True),
    ("I do not.", "I do.", False),
    ("She doesn't.", "She does.", True),
    ("She doesn't.", "She does.", False),
    ("She does not.", "She does.", True),
    ("She does not.", "She does.", False),
    ("She doesn't.", "She does.", True),
    ("She doesn't.", "She does.", False),
    ("She does not.", "She does.", True),
    ("She does not.", "She does.", False),
    ("I shouldn't.", "I should.", True),
    ("I shouldn't.", "I should.", False),
    ("I should not.", "I should.", True),
    ("I should not.", "I should.", False),
    ("I can't.", "I can.", True),
    ("I can't.", "I can.", False),
    ("I cannot.", "I can.", True),
    ("I cannot.", "I can.", False),
    ("I won't.", "I will.", True),
    ("I won't.", "I will.", False),
    ("I will not.", "I will.", True),
    ("I will not.", "I will.", False),
    ("I'm not excited.", "I'm excited.", True),
    ("I'm not excited.", "I'm excited.", False),
    ("I am not excited.", "I am excited.", True),
    ("I am not excited.", "I am excited.", False),
    ("I've not been excited.", "I've been excited.", True),
    ("I've not been excited.", "I've been excited.", False),
    ("I have not been excited.", "I have been excited.", True),
    ("I have not been excited.", "I have been excited.", False),
    ("I'll not be excited.", "I'll be excited.", True),
    ("I'll not be excited.", "I'll be excited.", False),
    ("I won't be excited.", "I will be excited.", True),
    ("I won't be excited.", "I will be excited.", False),
    ("I will not be excited.", "I will be excited.", True),
    ("I will not be excited.", "I will be excited.", False),
    ("He's not excited.", "He's excited.", True),
    ("He's not excited.", "He's excited.", False),
    ("He is not excited.", "He is excited.", True),
    ("He is not excited.", "He is excited.", False),
    ("We're not excited.", "We're excited.", True),
    ("We're not excited.", "We're excited.", False),
    ("We are not excited.", "We are excited.", True),
    ("We are not excited.", "We are excited.", False),
    ("You're not excited.", "You're excited.", True),
    ("You're not excited.", "You're excited.", False),
    ("You are not excited.", "You are excited.", True),
    ("You are not excited.", "You are excited.", False),
    ("They're not excited.", "They're excited.", True),
    ("They're not excited.", "They're excited.", False),
    ("They are not excited.", "They are excited.", True),
    ("They are not excited.", "They are excited.", False),
]

# AUX in ROOT children - Affirmative
aux_root_children_affirmative = [
    ("I do think it's true.", "I don't think it's true.", True),
    ("I do think it's true.", "I do not think it's true.", False),
    ("She does think it's true.", "She doesn't think it's true.", True),
    ("She does think it's true.", "She does not think it's true.", False),
    ("I should do it.", "I shouldn't do it.", True),
    ("I should do it.", "I should not do it.", False),
    ("You ought to do it.", "You ought not to do it.", False),
    ("I can do it.", "I can't do it.", True),
    ("I can do it.", "I cannot do it.", False),
    ("I will do it.", "I won't do it.", True),
    ("I will do it.", "I will not do it.", False),
    ("I've done it.", "I haven't done it.", True),
    ("I've done it.", "I have not done it.", False),
    ("I've been doing it.", "I haven't been doing it.", True),
    ("I've been doing it.", "I have not been doing it.", False),
    ("I have done it.", "I haven't done it.", True),
    ("I have done it.", "I have not done it.", False),
    ("I have been doing it.", "I haven't been doing it.", True),
    ("I have been doing it.", "I have not been doing it.", False),
    ("I'd done it.", "I hadn't done it.", True),
    ("I'd done it.", "I had not done it.", False),
    ("I had done it.", "I hadn't done it.", True),
    ("I had done it.", "I had not done it.", False),
    ("I had been doing it.", "I hadn't been doing it.", True),
    ("I had been doing it.", "I had not been doing it.", False),
    ("She's done it.", "She hasn't done it.", True),
    ("She's done it.", "She has not done it.", False),
    ("She's been doing it.", "She hasn't been doing it.", True),
    ("She's been doing it.", "She has not been doing it.", False),
    ("She has done it.", "She hasn't done it.", True),
    ("She has done it.", "She has not done it.", False),
    ("She has been doing it.", "She hasn't been doing it.", True),
    ("She has been doing it.", "She has not been doing it.", False),
    ("She'd done it.", "She hadn't done it.", True),
    ("She'd done it.", "She had not done it.", False),
    ("She'd been doing it.", "She hadn't been doing it.", True),
    ("She'd been doing it.", "She had not been doing it.", False),
    ("She had done it.", "She hadn't done it.", True),
    ("She had done it.", "She had not done it.", False),
    ("She had been doing it.", "She hadn't been doing it.", True),
    ("She had been doing it.", "She had not been doing it.", False),
    ("I would have done it.", "I wouldn't have done it.", True),
    ("I would have done it.", "I would not have done it.", False),
    ("I would've done it.", "I wouldn't have done it.", True),
    ("I would've done it.", "I would not have done it.", False),
    ("I should have done it.", "I shouldn't have done it.", True),
    ("I should have done it.", "I should not have done it.", False),
    ("I should've done it.", "I shouldn't have done it.", True),
    ("I should've done it.", "I should not have done it.", False),
]

# AUX in ROOT children - Negative
aux_root_children_negative = [
    ("I shouldn't do it.", "I should do it.", True),
    ("I shouldn't do it.", "I should do it.", False),
    ("I should not do it.", "I should do it.", True),
    ("I should not do it.", "I should do it.", False),
    ("I can't do it.", "I can do it.", True),
    ("I can't do it.", "I can do it.", False),
    ("I cannot do it.", "I can do it.", True),
    ("I cannot do it.", "I can do it.", False),
    ("I won't do it.", "I will do it.", True),
    ("I won't do it.", "I will do it.", False),
    ("I will not do it.", "I will do it.", True),
    ("I will no do it.", "I will do it.", False),
    ("I've not done it.", "I've done it.", True),
    ("I've not done it.", "I've done it.", False),
    ("I've not been doing it.", "I've been doing it.", True),
    ("I've not been doing it.", "I've been doing it.", False),
    ("I haven't done it.", "I have done it.", True),
    ("I haven't done it.", "I have done it.", False),
    ("I haven't been doing it.", "I have been doing it.", True),
    ("I haven't been doing it.", "I have been doing it.", False),
    ("I have not done it.", "I have done it.", True),
    ("I have not done it.", "I have done it.", False),
    ("I have not been doing it.", "I have been doing it.", True),
    ("I have not been doing it.", "I have been doing it.", False),
    ("I'd not done it.", "I'd done it.", True),
    ("I'd not done it.", "I'd done it.", False),
    ("I'd not been doing it.", "I'd been doing it.", True),
    ("I'd not been doing it.", "I'd been doing it.", False),
    ("I hadn't done it.", "I had done it.", True),
    ("I hadn't done it.", "I had done it.", False),
    ("I hadn't been doing it.", "I had been doing it.", True),
    ("I hadn't been doing it.", "I had been doing it.", False),
    ("I had not done it.", "I had done it.", True),
    ("I had not done it.", "I had done it.", False),
    ("I had not been doing it.", "I had been doing it.", True),
    ("I had not been doing it.", "I had been doing it.", False),
    ("She's not done it.", "She's done it.", True),
    ("She's not done it.", "She's done it.", False),
    ("She's not been doing it.", "She's been doing it.", True),
    ("She's not been doing it.", "She's been doing it.", False),
    ("She hasn't done it.", "She has done it.", True),
    ("She hasn't done it.", "She has done it.", False),
    ("She hasn't been doing it.", "She has been doing it.", True),
    ("She hasn't been doing it.", "She has been doing it.", False),
    ("She has not done it.", "She has done it.", True),
    ("She has not done it.", "She has done it.", False),
    ("She has not been doing it.", "She has been doing it.", True),
    ("She has not been doing it.", "She has been doing it.", False),
    ("I wouldn't have done it.", "I would have done it.", True),
    ("I wouldn't have done it.", "I would have done it.", False),
    ("I would not have done it.", "I would have done it.", True),
    ("I would not have done it.", "I would have done it.", False),
    ("I shouldn't have done it.", "I should have done it.", True),
    ("I shouldn't have done it.", "I should have done it.", False),
    ("I should not have done it.", "I should have done it.", True),
    ("I should not have done it.", "I should have done it.", False),
]

# General verbs - Affirmative
general_verbs_affirmative = [
    ("I love hiking.", "I don't love hiking.", True),
    ("I love hiking.", "I do not love hiking.", False),
    ("He loves hiking.", "He doesn't love hiking.", True),
    ("He loves hiking.", "He does not love hiking.", False),
    ("I used to love hiking.", "I didn't use to love hiking.", True),
    ("I used to love hiking.", "I did not use to love hiking.", False),
    ("I really liked the food.", "I really didn't like the food.", True),
    ("I really liked the food.", "I really did not like the food.", False),
]

# General verbs - Negative
general_verbs_negative = [
    ("I don't think it's true.", "I think it's true.", True),
    ("I don't think it's true.", "I think it's true.", False),
    ("I do not think it's true.", "I think it's true.", True),
    ("I do not think it's true.", "I think it's true.", False),
    ("She doesn't think it's true.", "She thinks it's true.", True),
    ("She doesn't think it's true.", "She thinks it's true.", False),
    ("She does not think it's true.", "She thinks it's true.", True),
    ("She does not think it's true.", "She thinks it's true.", False),
]

# Miscellaneous
misc = [
    ("A small Python module that doesn't negate sentences.", "A small Python module that negates sentences.", True),
    ("You should always be careful.", "You shouldn't always be careful.", True),
    ("You must be careful.", "You must not be careful.", False),
    ("I would have done it differently.", "I wouldn't have done it differently.", True),
    ("They are thinking too much and things might not go right.", "They aren't thinking too much and things might not go right.", True),
    ("They aren't going too fast and they might not have time.", "They are going too fast and they might not have time.", True),
    ("I am tired.", "I am not tired.", False),
    ("This was incorrectly assumed.", "This was not incorrectly assumed.", False),
    ("They were brought up in a small town.", "They weren't brought up in a small town.", True),
    ("I do like it.", "I don't like it.", True),
    ("She does like it.", "She doesn't like it.", True),
    ("He does like it.", "He does not like it.", False),
    ("He likes it.", "He doesn't like it.", True),
    ("He doesn't like it.", "He likes it.", True),
    ("I know everything.", "I don't know everything.", True),
    ("I don't know everything.", "I know everything.", True),
    ("The latter's chariot's wheels sank into the ground.", "The latter's chariot's wheels didn't sink into the ground.", True),
    ("He prepared a delicious meal.", "He didn't prepare a delicious meal.", True),
    ("I used to know everything.", "I did not use to know everything.", False),
    ("I'm so damn hungry.", "I'm not so damn hungry.", True),
    ("I am so damn hungry.", "I am not so damn hungry.", True),
    ("I am so damn hungry.", "I am not so damn hungry.", False),
    ("He's so afraid.", "He isn't so afraid.", True),
    ("He's so afraid.", "He is not so afraid.", False),
    ("They are kinda crazy.", "They are not kinda crazy.", False),
    ("There are many ways it can be done.", "There are not many ways it can be done.", False),
    ("There are many ways it can be done.", "There aren't many ways it can be done.", True),
    ("There exist many ways it can be done.", "There don't exist many ways it can be done.", True),
    ("For more info, visit our webpage.", "For more info, don't visit our webpage.", True),
    ("To dream, one doesn't need to sleep.", "To dream, one needs to sleep.", True),
    ("Make an iterator that returns object over and over again.", "Do not make an iterator that returns object over and over again.", False),
    ("Make an iterator that returns object over and over again.", "Don't make an iterator that returns object over and over again.", True),
    ("Do that again.", "Do not do that again.", False),
    ("Do that again.", "Don't do that again.", True),
    ("They will do it asap.", "They won't do it asap.", True),
    ("Not that this will ever change.", "Not that this won't ever change.", True),
    ("Not that this will ever change.", "Not that this will not ever change.", False),
    ("Not that this will not ever change.", "Not that this will ever change.", False),
    ("Not that this won't ever change.", "Not that this will ever change.", False),
    ("Not that he thinks a lot.", "Not that he doesn't think a lot.", True),
    ("Not that he thinks a lot.", "Not that he does not think a lot.", False),
    ("Not that he doesn't think a lot.", "Not that he thinks a lot.", False),
    ("Not that he does not think a lot.", "Not that he thinks a lot.", False),
    ("Not for the first time, she felt utterly betrayed.", "Not for the first time, she didn't feel utterly betrayed.", True),
    ("I've studied hard.", "I haven't studied hard.", True),
    ("I've studied hard.", "I have not studied hard.", False),
    ("She's been to Paris.", "She has not been to Paris.", False),
    ("She's done it.", "She hasn't done it.", True),
    ("She's running.", "She isn't running.", True),
    ("He's rather shy.", "He isn't rather shy.", True),
    ("He's Paul.", "He is not Paul.", False),
    ("I will never go there.", "I will go there.", False),
    ("I always go there.", "I always don't go there.", True)  # not very natural
]

# Sentences that are currently failing.
failing = [
    # POS tagger misclassifications.
    ("I do.", "I don't.", True),  # "do" classified as VERB instead of AUX.
    ("I do.", "I do not.", False),  # "do" classified as VERB instead of AUX.
    ("She does.", "She doesn't.", True),  # "do" classified as VERB instead of AUX.
    ("She does.", "She does not.", False),  # "do" classified as VERB instead of AUX.
    ("She's determined.", "She is not determined.", False),  # "determined" classified as VERB instead of ADJ.
    # Connectors.
    ("It also prohibits or restricts the use of certain mechanisms.", "It also doesn't prohibit or restrict the use of certain mechanisms.", True),
    # Boolean logic, e.g.: "prohibits AND restricts" -> "doesn't prohibit OR restrict"
    ("It also prohibits and restricts the use of certain mechanisms.", "It also doesn't prohibit or restrict the use of certain mechanisms.", True),
    # Non-verbal negations.
    ("A piece with no moving parts.", "A piece with moving parts.", False),  # No verb to negate.
    # Special case of negated ought not yet implemented.
    ("You ought not to do it.", "You ought to do it.", False),
    ("You ought to do it.", "You oughtn't do it.", True),
    ("You oughtn't do it.", "You ought to do it.", True),
]



"""Tests for sentence negation."""

import pytest
from negate import Negator
from .data import (aux_root_affirmative, aux_root_negative,
                   aux_root_children_affirmative, aux_root_children_negative,
                   general_verbs_affirmative, general_verbs_negative,
                   misc, failing)


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    aux_root_affirmative
)
def test_aux_root_affirmative(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    aux_root_negative
)
def test_aux_root_negative(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    aux_root_children_affirmative
)
def test_aux_root_children_affirmative(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    aux_root_children_negative
)
def test_aux_root_children_negative(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    general_verbs_affirmative
)
def test_general_verbs_affirmative(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    general_verbs_negative
)
def test_general_verbs_negative(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence


@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    misc
)
def test_misc(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence

@pytest.mark.xfail
@pytest.mark.parametrize(
    "input_sentence, output_sentence, prefer_contractions",
    failing
)
def test_failing(
    negator: Negator,  # pylint: disable=redefined-outer-name
    input_sentence: str,
    output_sentence: str,
    prefer_contractions: bool
):
    assert negator.negate_sentence(
        input_sentence, prefer_contractions) == output_sentence



"""Token-related utils."""

from dataclasses import dataclass


@dataclass
class Token:
    """A token, i.e., a sequence of characters.

    A :obj:`spacy.tokens.Token` is read-only. See
    `https://spacy.io/api/token`__. Furthermore, it cannot hold an empty string
    as :attr:`text`. This class is just a minimalist equivalent to a Spacy Token
    that is both writable and allows empty strings as text. It only contains the
    attributes relevant for our purposes.

    Attributes:
        text (:obj:`str`):
            The characters that form the token.
        has_space_after (:obj:`str`):
            Whether the token is followed by a whitespace (this whitespace is
            not part of the :attr:`text`). This information is used for
            detokenization.
    """

    text: str
    has_space_after: bool = True
